# -*- coding: utf-8 -*-

# Copyright: © 2013 Olivier Larchevêque <olivier.larcheveque@auf.org>
#
# This file is part of auf-piwik.
#
# auf-piwik is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# auf-piwik is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with auf-piwik.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import urllib
import StringIO
import datetime

import import_logs

from import_logs import *  # noqa
from settings import Settings, mod_logging

piwik_settings = Settings()


class AUFStatistics(Statistics):
    """
    stdout => log info
    """
    def print_summary(self):
        stdold, stdnew = sys.stdout, StringIO.StringIO()
        sys.stdout = stdnew
        super(AUFStatistics, self).print_summary()
        sys.stdout = stdold
        summary = stdnew.getvalue()
        logging.info(summary)
        return summary


class AUFConfiguration(Configuration):
    """
    """
    def _create_parser(self):
        global logging
        option_parser = super(AUFConfiguration, self)._create_parser()
        option_parser.add_option(
            '--logging', dest='logging_level', default='ERROR',
            help=u"Spécifie le niveau de log",
        )
        option_parser.add_option(
            '--site-name', dest='site_name', default='Pas de dénomination',
            help=u"Spécifie de quel site on parle (pour logging)",
        )
        option_parser.add_option(
            '--accesslog', dest='accesslog',
            help=u"Path du fichier access.log du site"
        )
        options, filenames = option_parser.parse_args(sys.argv[1:])
        logging = piwik_settings.get_logger(options.site_name)
        logging.setLevel(getattr(mod_logging, options.logging_level))
        import_logs.logging = logging

        if os.path.exists(options.accesslog):
            sys.argv.append(options.accesslog)
        else:
            raise Exception(u"Le fichier de logs est introuvable: %s" %
                            options.accesslog)

        sys.argv.append("%s.1" % options.accesslog)

        # Monkey patch pour avoir le me système de log partout
        def fatal_error(error, filename=None, lineno=None):
            logging.error("[%s:%s] %s " % (filename, lineno, error))
            sys.exit(-1)
        import_logs.fatal_error = fatal_error

        return option_parser

    def _parse_args(self, option_parser):
        """
        Surcharge pour retirer la configuration du logger.
        """
        self.options, self.filenames = option_parser.parse_args(sys.argv[1:])

        if self.options.output:
            sys.stdout = sys.stderr = open(self.options.output, 'a+', 0)

        if not self.filenames:
            print(option_parser.format_help())
            sys.exit(1)

        # Configure logging before calling logging.{debug,info}.
        #logging.basicConfig(
        #    format='%(asctime)s: [%(levelname)s] %(message)s',
        #    level=logging.DEBUG if self.options.debug >= 1 else logging.INFO,
        #)

        self.options.excluded_useragents = \
            [s.lower() for s in self.options.excluded_useragents]

        if self.options.exclude_path_from:
            paths = [path.strip() for path
                     in open(self.options.exclude_path_from).readlines()]
            self.options.excluded_paths.extend(
                path for path in paths if len(path) > 0)
        if self.options.excluded_paths:
            logging.debug('Excluded paths: %s',
                          ' '.join(self.options.excluded_paths))

        if self.options.hostnames:
            logging.debug('Accepted hostnames: %s',
                          ', '.join(self.options.hostnames))
        else:
            logging.debug('Accepted hostnames: all')

        if self.options.log_format_regex:
            self.format = RegexFormat('custom', self.options.log_format_regex)
        elif self.options.log_format_name:
            try:
                self.format = FORMATS[self.options.log_format_name]
            except KeyError:
                fatal_error('invalid log format: %s' %
                            self.options.log_format_name)
        else:
            self.format = None

        if not self.options.piwik_url:
            fatal_error('no URL given for Piwik')

        if not (self.options.piwik_url.startswith('http://') or
                self.options.piwik_url.startswith('https://')):
            self.options.piwik_url = 'http://' + self.options.piwik_url
        logging.debug('Piwik URL is: %s', self.options.piwik_url)

        if not self.options.piwik_token_auth:
            try:
                self.options.piwik_token_auth = self._get_token_auth()
            except Piwik.Error, e:
                fatal_error(e)
        logging.debug('Authentication token token_auth is: %s',
                      self.options.piwik_token_auth)

        if self.options.recorders < 1:
            self.options.recorders = 1


class AUFParser(Parser):
    """
    Surcharge du parser pour ne pas réimporter des logs déjà traités.
    """
    idsite = None
    date = None
    date_format = "%Y-%m-%d %H:%M:%S"
    delta_start_days = 365
    look_backward = "last%s" % delta_start_days
    max_date = None

    def __init__(self):
        super(AUFParser, self).__init__()
        self.site_name = config.options.site_name
        self.idsite = config.options.site_id
        filename = self.get_import_file()
        if not os.path.exists(filename):
            open(filename, 'w').close()
            logging.debug(u"création du fichier %s" % filename)

        self.sous_repertoire = None
        url_segments = urlparse.urlparse(self.site_name)
        path = url_segments[2].strip('/').lower()
        if path:
            self.sous_repertoire = path
        logging.debug(u"sous-repertoire: %s" % self.sous_repertoire)

    def get_import_file(self):
        return os.path.join(piwik_settings.piwik_cache, self.idsite)

    def get_last_piwik_tracking(self):
        """
        On regarde la dernière info d'importation (dans une certaine limite de
        temps).
        Attention, on ajoute 1 jour car l'API ne totalise pas le jour courant.
        """
        params = {
            'module': 'API',
            'method': 'Actions.get',
            'idSite': config.options.site_id,
            'date': self.look_backward,
            'period': 'day',
            'format': 'json',
            'token_auth': config.options.piwik_token_auth,
            }
        params = urllib.urlencode(params)
        url = "%s?%s" % (config.options.piwik_url, params)
        json_data = urllib.urlopen(url).read()
        logging.debug(u"Piwik get stats: %s" % url)

        results = json.load(StringIO.StringIO(json_data))
        track_dates = [k for k, v in results.items() if len(v) > 0]
        if len(track_dates) > 0:
            logging.debug(u"dernière date récupérée de piwik")
            date = datetime.datetime.strptime(track_dates[0], "%Y-%m-%d") + \
                datetime.timedelta(days=1)
            return date
        else:
            logging.debug(u"dernière date fixée à %s jours avants" %
                          self.delta_start_days)
            return datetime.datetime.now() - \
                datetime.timedelta(days=self.delta_start_days)

    def get_last_import_date(self):
        """
        On cherche la date du dernier import, autrement on le définit à une
        année avant.
        """
        if self.date:
            return self.date

        filename = self.get_import_file()
        f = open(filename, 'r')
        try:
            str_date = f.read().replace('\n', '')
            date = datetime.datetime.strptime(str_date, self.date_format)
            logging.debug(u"date lue du cache: %s" % date)
        except:
            date = self.get_last_piwik_tracking()
            logging.debug(u"initialisation de la date à: %s" % date)
            self.set_last_import_date(date=date)
            date += datetime.timedelta(days=1)
        f.close()
        logging.debug(u"lecture de la dernière date d'importation: %s" %
                      date)
        self.date = date
        return date

    def set_last_import_date(self, date=None):
        if date is None:
            date = self.max_date
        if not config.options.dry_run:
            f = open(self.get_import_file(), 'w')
            f.write(date.strftime(self.date_format))
            f.close()
        logging.debug(u"écriture de la dernière date d'importation: %s" % date)

    def check_last_import_performed(self, hit):
        """
        Si on a pas encore la date dans le hit, on s'y intéresse dans tous les
        cas.
        On compare sa date avec la date du dernier import, s'il le hit a été
        fait sur Piwik (pas de cache), après le dernier import + 1 jour
        (car on n'a pas les secondes), il sera ajouté à la liste.
        """
        if not hasattr(hit, 'date'):
            return True
        if self.max_date is None or hit.date > self.max_date:
            self.max_date = hit.date
        available_for_import = self.get_last_import_date()
        do_import = hit.date > available_for_import
        return do_import

    def check_path_sous_repertoire(self, hit):
        """
        Si le site est dans un sous-répertoire, on élimine les hits qui ne sont
        pas préfixé de la sorte.
        """
        if not hasattr(hit, 'path'):
            return True
        if self.sous_repertoire is None:
            return True
        path = hit.path.lower().strip('/')

        if path.startswith(self.sous_repertoire):
            return True
        else:
            return False


class AUFPiwik(Piwik):

    @staticmethod
    def _call(path, args, headers=None, url=None, data=None):
        """
        Surcharge de la méthode car le fonctionnement via POST requiert une
        token SuperUser.
        La fonction retraite les datas pour qu'elles soient dans le format
        attendu par l'API de tracking:
        (http://piwik.org/docs/tracking-api/
        reference/#toc-advanced-bulk-tracking-requests)

        {
           "requests": [
              "?idsite=1&url=http://ex.org&action_name=Test bulk Page&rec=1",
              "?idsite=1&url=http://ex.net/test.htm&action_name=Add bul&rec=1"
           ],
           "token_auth": "33dc3f2536d3025974cccb4b4d2d98f4"
        }
        """
        logging.debug("API _call")
        logging.debug(path)
        logging.debug(headers)
        logging.debug(url)
        #logging.debug(data)

        if data is None:
            return _call(path, args, headers, url, data)

        if url is None:
            url = config.options.piwik_url
        headers = headers or {}

        url_requests = []
        for dic in data['requests']:
            try:
                params = urllib.urlencode([(k, v) for k, v in dic.items()])
                url_requests.append("?%s" % params)
            except:
                logging.debug("requêtes API piwik: %s" % dic)

        data['requests'] = url_requests
        data = json.dumps(data)

        request = urllib2.Request(url + path, data, headers)
        response = urllib2.urlopen(request)
        result = response.read()
        response.close()
        return result


if __name__ == '__main__':
    try:
        piwik = AUFPiwik()
        import_logs.piwik = piwik

        config = AUFConfiguration()
        import_logs.logging = logging
        config.options.show_progress = False
        import_logs.config = config

        stats = AUFStatistics()
        import_logs.stats = stats

        resolver = config.get_resolver()
        import_logs.resolver = resolver

        parser = AUFParser()
        import_logs.parser = parser

        main()
        parser.set_last_import_date()
    except Exception, e:
        logging.exception(e)
        sys.exit(-1)
    sys.exit(0)
